header {* experiments on scanning and parsing
          following Stefan Berghofers presentation
          at the Isabelle Developer Workshop 2010 *}

theory Test_Parsers
imports Main
begin

text {* parse following Stefan Berghofer at Isabelle Developer Workshop 2010 *}
ML {* 
fun filtered_input str =
  filter Token.is_proper (Outer_Syntax.scan Position.none str);
fun parse p str = Scan.finite Token.stopper p (filtered_input str);
*}

section {* scanner combinators from src/Pure/General/scan.ML *}
text {* ATTENTION, there are several similar definitions:
src> grep -r 'infix ' * | grep '\-\- ' | less
HOL/Import/proof_kernel.ML:    infix 7 |-- --|
HOL/Import/proof_kernel.ML:    infix 5 :-- -- ^^
HOL/Import/mono_scan.ML:infix 7 |-- --|
HOL/Import/mono_scan.ML:infix 5 :-- -- ^^
HOL/Import/scan.ML:infix 7 |-- --|
HOL/Import/scan.ML:infix 5 :-- -- ^^
Pure/General/scan.ML:infix 5 -- :-- :|-- |-- --| ^^;
*}

subsection {* one typical usage per combinator *}
text {* typical signatures *}
ML {*
parse;
(*fn : (Token.T list -> 'a * Token.T list) -> string -> 'a * Token.T list*)
      Parse.nat;
      (*fn : int Parse.parser*)
      (*fn : (Token.T list -> 'a * Token.T list)*)
parse Parse.nat;
(*fn : string -> int * Token.T list*)

parse Parse.nat "123";
(*(123, []) : int * Token.T list*)
*}

subsubsection {* op >> *}
ML {*
Parse.nat;
(fn i => i + 1);
op >>;
(Parse.nat >> (fn i => i + 1));
parse;
parse (Parse.nat >> (fn i => i + 1));
(*
parse (Parse.nat >> (fn i => i + 1));
       int Parse.parser
                     int -> int
                 ('a -> 'b * 'c) * ('b -> 'd) -> 'a -> 'd * 'c
      Token.T list -> int * Token.T list
(Token.T list -> 'a * Token.T list) -> string -> 'a * Token.T list
val it = fn : string -> int * Token.T list
*)
*}

subsubsection {* op || *}
ML {*
parse (Args.$$$ "aaa" || Args.$$$ "bbb" || Args.$$$ "ccc") "aaa 111";
parse (Args.$$$ "aaa" || Args.$$$ "bbb" || Args.$$$ "ccc") "ccc 111";
(*val it = ("ccc", [Token (("111", ({}, {})), (Nat, "111"), Slot)])
   : string * Token.T list*)
parse ((Args.$$$ "aaa" -- Parse.nat) || (Args.$$$ "bbb" -- Parse.nat)) "aaa 111";
parse ((Args.$$$ "aaa" -- Parse.nat) || (Args.$$$ "bbb" -- Parse.nat)) "bbb 222";
(*val it = (("bbb", 222), []) : (string * int) * Token.T list*)
(* ?TODO how || with more different alternatives?
parse (Args.name || Parse.nat) "aaa 111 bbb 222";
***    Argument: (Args.name, Parse.nat) : string parser * int Parse.parser
***    Reason:
***    Can't unify string (*In Basis*) with int (*In Basis*)
***       (Different type constructors)*)
*}

subsubsection {* op :-- *}
ML {* (*from src/Pure/Isar/*)
            Parse.sym_ident : Token.T list -> string * Token.T list;
            (*fn : string Parse.parser*)
                                (fn "-" => Scan.succeed () | _ => Scan.fail);
                                (fn "-" => Scan.succeed ());
                                (*fn : string -> 'a -> unit * 'a*)
            Parse.sym_ident :-- (fn "-" => Scan.succeed () | _ => Scan.fail);
            (*fn : Token.T list -> (string * unit) * Token.T list*)
op >>;
(*fn : ('a -> 'b * 'c) * ('b -> 'd) -> 'a -> 'd * 'c*)

val minus = Parse.sym_ident :-- (fn "-" => Scan.succeed () | _ => Scan.fail) >> #1;
(*fn : Token.T list -> string * Token.T list*)
"------------------------------------------";
parse minus "-aaa";
parse minus "- -";
(*WHY NOT SIMPLY Args.$$$ "-" ?*)
(*parse minus "111 -";*)
"------------------------------------------";
infix 5 :--;
fun (scan1 -- scan2) = scan1 :-- (fn _ => scan2);
*}

subsubsection {* op -- *}
ML {*
parse (minus -- Parse.nat)  "- 111";

Args.$$$ "theory";
Args.name;
(Args.$$$ "theory" -- Args.name);
parse (Args.$$$ "theory" -- Args.name);
(*
parse (Args.$$$ "theory")                GOON ?!?!?
  "theory";

parse (Args.$$$ "theory")
  "theory Sketch imports Main";
parse (Args.$$$ "theory" -- Args.name)
  "theory Sketch imports Main";
parse (Args.$$$ "theory" -- Args.name -- Args.$$$ "imports" -- Args.name)
  "theory Sketch imports Main";
*)
*}

subsubsection {* op :|-- TODO*}
ML {*
fun token atom = Scan.ahead Parse.not_eof --| atom;
(*fn : (Token.T list -> 'a * 'b) -> Token.T list -> Token.T * 'b*)

val ident = token
  (Parse.short_ident || Parse.long_ident || Parse.sym_ident || Parse.term_var ||
    Parse.type_ident || Parse.type_var || Parse.number);
(*fn : Token.T list -> Token.T * Token.T list*)

(ident >> Token.content_of) :|-- (fn x =>
  (case Lexicon.read_variable x of SOME v => Scan.succeed v | NONE => Scan.fail));
(*fn : Token.T list -> indexname * Token.T list*)
*}

subsubsection {* op |-- *}
ML {* (*from Pure/Isar/args.ML, fun attribs*)
parse (Parse.$$$ "[" |-- Parse.list1 Parse.name) "[111, bbb, ccc]";
*}

subsubsection {* op --| *}
ML {* (*from Pure/Isar/args.ML, fun attribs*)
parse (Parse.list1 Parse.name --| Parse.$$$ "]") " 111, bbb, ccc]";
*}

section {* parse compound suffixes *}
subsection {* Lists *}
ML {* (*from Pure/Isar/args.ML, fun attribs*)
Parse.$$$ "[" |-- Parse.list1 Parse.name --| Parse.$$$ "]";
(*fn : Token.T list -> bstring list * Token.T list*)

parse (Parse.$$$ "[" |-- Parse.list Parse.name --| Parse.$$$ "]") "[aaa, bbb, ccc]"; 
*}

subsection {* terms *}
text {* see src/Pure/isac/smltest/Pure/Isar/Test_Parse_Term.thy;
        plus (1), (2) below*}
text {*(1) why repeat required ?*}
ML {*
Parse.term;
parse Parse.term;
parse Parse.term "xxx +++ 111 === 222";
(*val it =
   ("\^E\^Ftoken\^Exxx\^E\^F\^E",
    [Token (("+++", ({}, {})), (SymIdent, "+++"), Slot),
     Token (("111", ({}, {})), (Nat, "111"), Slot),
     Token (("===", ({}, {})), (SymIdent, "==="), Slot),
     Token (("222", ({}, {})), (Nat, "222"), Slot)])
   : string * Token.T list*)
*}

text {*(2) +( are keywords in outer/inner syntax ?
       what is Token.kind ... ?*}
ML {*
"---1 OK--------------------------------------------------";
parse (Scan.repeat Parse.term) "solve xxx +++ 111 === 222 end";
"---2 NOK by '+' -----------------------------------------";
parse (Scan.repeat Parse.term) "xxx + 111 === 222 end";
"---3 NOK by '(' -----------------------------------------";
parse (Scan.repeat Parse.term) "solve (xxx +++ 111 === 222) end";
*}

text {* decompose following "assume" --> Parse_Spec.statement*}
ML {* (*from Pure/Isar/isar_syn.ML*)
Outer_Syntax.command "assume" "assume propositions"
  (Keyword.tag_proof Keyword.prf_asm)
    (Parse_Spec.statement >> (Toplevel.print oo (Toplevel.proof o Proof.assume_cmd)));
(*() : unit*)
Outer_Syntax.command "assume" "assume propositions";
(*fn : Keyword.T -> (Toplevel.transition -> Toplevel.transition) parser -> unit*)

  (Keyword.tag_proof Keyword.prf_asm);
  (*Keyword ("proof-asm", ["proof"]) : Keyword.T*)

Outer_Syntax.command "assume" "assume propositions" (Keyword.tag_proof Keyword.prf_asm);
(*fn : (Toplevel.transition -> Toplevel.transition) parser -> unit*)

    (Parse_Spec.statement >> (Toplevel.print oo (Toplevel.proof o Proof.assume_cmd)));
    (*fn : Token.T list -> (Toplevel.transition -> Toplevel.transition) * Token.T list*)

          (*val statement = Parse.and_list1 (opt_thm_name ":" -- Scan.repeat1 Parse.propp);*)
                                                                 Scan.repeat1 Parse.propp;

parse (Scan.repeat1 Parse.propp) "xxx +++  111  === 222 end";
(*val it =
   ([("\^E\^Ftoken\^Exxx\^E\^F\^E", []), ("\^E\^Ftoken\^E+++\^E\^F\^E", []),
     ("\^E\^Ftoken\^E111\^E\^F\^E", []), ("\^E\^Ftoken\^E===\^E\^F\^E", []),
     ("\^E\^Ftoken\^E222\^E\^F\^E", [])],
    [Token (("end", ({}, {})), (Command, "end"), Slot)])
   : (string * string list) list * Token.T list*)
*}



section {* GOON *}
subsection {* GOON *}


end